from typing import Optional

from slixmpp import JID, Iq

from slidge import BaseGateway, BaseSession, GatewayUser
from slidge.contact import LegacyContact
from slidge.util.test import SlidgeTest
from slidge.util.types import RecipientType, LegacyMessageType, LegacyThreadType


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"


class Session(BaseSession):
    async def login(self):
        pass

    async def on_react(
        self,
        chat: RecipientType,
        legacy_msg_id: LegacyMessageType,
        emojis: list[str],
        thread: Optional[LegacyThreadType] = None,
    ):
        pass

class Contact(LegacyContact):
    REACTIONS_SINGLE_EMOJI = True


class TestGatewayWideRestrictions(SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="romeo@montague.lit/gajim")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "logged in" in stanza["status"].lower(), stanza

        self.send(  # language=XML
            """
            <iq type="get"
                to="romeo@montague.lit"
                id="1"
                from="aim.shakespeare.lit">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:avatar:metadata" />
              </pubsub>
            </iq>
            """
        )

    @property
    def romeo_session(self) -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    def test_gateway_wide_restrictions(self):
        self.recv(  # language=XML
            f"""
            <iq from='romeo@montague.lit'
                to='{self.xmpp.boundjid.bare}'
                type='get'>
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                from="aim.shakespeare.lit"
                to="romeo@montague.lit"
                type="result"
                id="2">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="account"
                          type="registered"
                          name="A test" />
                <identity category="pubsub"
                          type="pep"
                          name="A test" />
                <identity category="gateway"
                          type=""
                          name="A test" />
                <feature var="jabber:iq:search" />
                <feature var="jabber:iq:register" />
                <feature var="urn:ietf:params:xml:ns:vcard-4.0" />
                <feature var="urn:xmpp:mds:displayed:0+notify" />
                <feature var="urn:xmpp:mds:displayed:0" />
                <feature var="http://jabber.org/protocol/pubsub#event" />
                <feature var="http://jabber.org/protocol/pubsub#retrieve-items" />
                <feature var="http://jabber.org/protocol/pubsub#persistent-items" />
                <feature var="urn:xmpp:avatar:metadata+notify" />
                <feature var="urn:xmpp:chat-markers:0" />
                <feature var="jabber:iq:gateway" />
                <feature var="urn:xmpp:ping" />
                <feature var="http://jabber.org/protocol/commands" />
                <x xmlns="jabber:x:data"
                   type="result">
                  <field var="FORM_TYPE"
                         type="hidden">
                    <value>urn:xmpp:reactions:0:restrictions</value>
                  </field>
                  <field var="max_reactions_per_user"
                         type="number">
                    <value>1</value>
                  </field>
                  <field var="scope">
                    <value>domain</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )


    def test_error_multi_emoji(self):
        # one of the other tests has a persisting side effect which is triggered only
        # when running the full suite. :(
        self.xmpp.plugin["xep_0356"].granted_privileges.clear()
        self.recv(  # language=XML
            f"""
            <message from='romeo@montague.lit'
                     to='juliet@{self.xmpp.boundjid.bare}'>
              <reactions id='some-id'
                         xmlns='urn:xmpp:reactions:0'>
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
            </message>
            """
        )
        self.send(  # language=XML
            f"""
            <message xmlns="jabber:component:accept"
                     type="chat"
                     from="aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <body>Maximum 1 emoji/message</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit"
                     type="error">
              <reactions xmlns="urn:xmpp:reactions:0"
                         id="some-id">
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
              <error type="modify">
                <feature-not-implemented xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Maximum 1 emoji/message</text>
              </error>
            </message>
            """,
            use_values=False
        )



from .util import (
    SubclassableOnce,
    is_valid_phone_number,
    replace_mentions,
    strip_illegal_chars,
)

__all__ = [
    "SubclassableOnce",
    "is_valid_phone_number",
    "replace_mentions",
    "strip_illegal_chars",
]



"""
Prettify XML in slixmpp tests, using utidylib
"""

import re
from argparse import ArgumentParser
from pathlib import Path
from textwrap import indent

import tidy


def get_parser():
    parser = ArgumentParser()
    parser.add_argument("INPUT", nargs="*", help="Python file or dir to process")
    return parser


def main():
    args = get_parser().parse_args()
    for path in args.INPUT:
        path = Path(path)
        if path.is_file():
            process(path)
        else:
            for file in path.glob("**/*.py"):
                process(file)


def process(path: Path):
    content = path.read_text()
    pretty = re.sub(PATTERN, sub, content)
    new = []
    # remove useless language injection tags and place them at the right place
    for line in pretty.split("\n"):
        stripped = line.strip()
        if stripped in ("self.recv(", "self.send("):
            new.append(line.replace("# language=XML", "").rstrip() + "  # language=XML")
        else:
            new.append(line)
    new = "\n".join(new)
    if new != content:
        print(f"Prettyfying {path}")
        path.write_text(new)


def sub(match: re.Match):
    whole = match.group(0)
    lines = whole.lstrip(" \n").split("\n")
    first = lines[0][0]
    if first == "#" or first != '"':
        return whole
    line = lines[1 if len(lines) > 1 else 0].lstrip(" \n")
    if not line:
        return whole
    first_non_quote = line[0]
    if first_non_quote != "<":
        return whole
    return '"""\n' + indent(prettify(match.group(1)) + '"""', " " * 12)


def prettify(xml: str):
    r = str(
        tidy.parseString(
            xml,
            input_xml=True,
            indent_attributes=True,
            indent="auto",
            indent_spaces=2,
            # we don't want to wrap anything because whitespace
            # in tag text is meaningful
            wrap=10_000,
        )
    )
    lines = r.split("\n")
    return "\n".join(line.rstrip() for line in lines)


def strip_first_line(s: str):
    return "\n".join(s.split("\n")[1:])


PATTERN = re.compile(r'"""(.*?)"""', re.MULTILINE | re.DOTALL)


if __name__ == "__main__":
    main()



import logging
import mimetypes
import re
import warnings
from collections.abc import Callable
from functools import wraps
from pathlib import Path
from time import time
from typing import Any, NamedTuple, TypeVar

try:
    import emoji
except ImportError:
    EMOJI_LIB_AVAILABLE = False
else:
    EMOJI_LIB_AVAILABLE = True

from slixmpp.types import ResourceDict

from .types import Mention

try:
    import magic
except ImportError as e:
    magic = None  # type:ignore
    logging.warning(
        (
            "Libmagic is not available: %s. "
            "It's OK if you don't use fix-filename-suffix-mime-type."
        ),
        e,
    )


def fix_suffix(
    path: Path, mime_type: str | None, file_name: str | None
) -> tuple[str, str]:
    guessed = magic.from_file(path, mime=True)
    if guessed == mime_type:
        log.debug("Magic and given MIME match")
    else:
        log.debug("Magic (%s) and given MIME (%s) differ", guessed, mime_type)
        mime_type = guessed

    valid_suffix_list = mimetypes.guess_all_extensions(mime_type, strict=False)

    if file_name:
        name = Path(file_name)
    else:
        name = Path(path.name)

    suffix = name.suffix

    if suffix in valid_suffix_list:
        log.debug("Suffix %s is in %s", suffix, valid_suffix_list)
        return str(name), guessed

    valid_suffix = mimetypes.guess_extension(mime_type.split(";")[0], strict=False)
    if valid_suffix is None:
        log.debug("No valid suffix found")
        return str(name), guessed

    log.debug("Changing suffix of %s to %s", file_name or path.name, valid_suffix)
    return str(name.with_suffix(valid_suffix)), guessed


class SubclassableOnce:
    # To allow importing everything, including plugins, during tests
    TEST_MODE: bool = False
    __subclasses: dict[type["SubclassableOnce"], type["SubclassableOnce"] | None] = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if SubclassableOnce not in cls.__bases__:
            base = SubclassableOnce.__find_direct_child(cls)
            existing = SubclassableOnce.__subclasses.get(base)
            if existing is not None and not SubclassableOnce.TEST_MODE:
                raise RuntimeError("This class must be subclassed once at most!")
            cls.__subclasses[base] = cls
        super().__init_subclass__(**kwargs)

    @staticmethod
    def __find_direct_child(cls: type["SubclassableOnce"]) -> type["SubclassableOnce"]:
        for base in cls.__bases__:
            if issubclass(base, SubclassableOnce):
                return base
        else:
            raise RuntimeError("wut")

    @classmethod
    def get_self_or_unique_subclass(cls) -> "type[SubclassableOnce]":
        try:
            return cls.get_unique_subclass()
        except AttributeError:
            return cls

    @classmethod
    def get_unique_subclass(cls) -> "type[SubclassableOnce]":
        existing = SubclassableOnce.__subclasses.get(cls)
        if existing is None:
            raise AttributeError("Could not find any subclass", cls)
        return existing

    @classmethod
    def reset_subclass(cls) -> None:
        log.debug("Resetting subclass of %s", cls)
        cls.__subclasses[cls] = None


def is_valid_phone_number(phone: str | None):
    if phone is None:
        return False
    match = re.match(r"\+\d.*", phone)
    if match is None:
        return False
    return match[0] == phone


def strip_illegal_chars(s: str, repl: str = "") -> str:
    return ILLEGAL_XML_CHARS_RE.sub(repl, s)


# from https://stackoverflow.com/a/64570125/5902284 and Link Mauve
ILLEGAL = [
    (0x00, 0x08),
    (0x0B, 0x0C),
    (0x0E, 0x1F),
    (0x7F, 0x84),
    (0x86, 0x9F),
    (0xFDD0, 0xFDDF),
    (0xFFFE, 0xFFFF),
    (0x1FFFE, 0x1FFFF),
    (0x2FFFE, 0x2FFFF),
    (0x3FFFE, 0x3FFFF),
    (0x4FFFE, 0x4FFFF),
    (0x5FFFE, 0x5FFFF),
    (0x6FFFE, 0x6FFFF),
    (0x7FFFE, 0x7FFFF),
    (0x8FFFE, 0x8FFFF),
    (0x9FFFE, 0x9FFFF),
    (0xAFFFE, 0xAFFFF),
    (0xBFFFE, 0xBFFFF),
    (0xCFFFE, 0xCFFFF),
    (0xDFFFE, 0xDFFFF),
    (0xEFFFE, 0xEFFFF),
    (0xFFFFE, 0xFFFFF),
    (0x10FFFE, 0x10FFFF),
]

ILLEGAL_RANGES = [rf"{chr(low)}-{chr(high)}" for (low, high) in ILLEGAL]
XML_ILLEGAL_CHARACTER_REGEX = "[" + "".join(ILLEGAL_RANGES) + "]"
ILLEGAL_XML_CHARS_RE = re.compile(XML_ILLEGAL_CHARACTER_REGEX)


# from https://stackoverflow.com/a/35804945/5902284
def addLoggingLevel(
    levelName: str = "TRACE", levelNum: int = logging.DEBUG - 5, methodName=None
) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        log.debug(f"{levelName} already defined in logging module")
        return
    if hasattr(logging, methodName):
        log.debug(f"{methodName} already defined in logging module")
        return
    if hasattr(logging.getLoggerClass(), methodName):
        log.debug(f"{methodName} already defined in logger class")
        return

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs) -> None:
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs) -> None:
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


class SlidgeLogger(logging.Logger):
    def trace(self) -> None:
        pass


log = logging.getLogger(__name__)


def merge_resources(resources: dict[str, ResourceDict]) -> ResourceDict | None:
    if len(resources) == 0:
        return None

    if len(resources) == 1:
        return next(iter(resources.values()))

    by_priority = sorted(resources.values(), key=lambda r: r["priority"], reverse=True)

    if any(r["show"] == "" for r in resources.values()):
        # if a client is "available", we're "available"
        show = ""
    else:
        for r in by_priority:
            if r["show"]:
                show = r["show"]
                break
        else:
            raise RuntimeError()

    # if there are different statuses, we use the highest priority one,
    # but we ignore resources without status, even with high priority
    status = ""
    for r in by_priority:
        if r["status"]:
            status = r["status"]
            break

    return {
        "show": show,  # type:ignore
        "status": status,
        "priority": 0,
    }


def remove_emoji_variation_selector_16(emoji: str):
    # this is required for compatibility with dino, and maybe other future clients?
    return bytes(emoji, encoding="utf-8").replace(b"\xef\xb8\x8f", b"").decode()


def deprecated(name: str, new: Callable):
    # @functools.wraps
    def wrapped(*args, **kwargs):
        warnings.warn(
            f"{name} is deprecated. Use {new.__name__} instead",
            category=DeprecationWarning,
        )
        return new(*args, **kwargs)

    return wrapped


T = TypeVar("T", bound=NamedTuple)


def dict_to_named_tuple(data: dict, cls: type[T]) -> T:
    return cls(*(data.get(f) for f in cls._fields))  # type:ignore


def replace_mentions(
    text: str,
    mentions: list[Mention] | None,
    mapping: Callable[[Mention], str],
):
    if not mentions:
        return text

    cursor = 0
    pieces = []
    for mention in mentions:
        try:
            new_text = mapping(mention)
        except Exception as exc:
            log.debug("Attempting slidge <= 0.3.3 compatibility: %s", exc)
            new_text = mapping(mention.contact)  # type:ignore
        pieces.extend([text[cursor : mention.start], new_text])
        cursor = mention.end
    pieces.append(text[cursor:])
    return "".join(pieces)


def timeit(func):
    @wraps(func)
    async def wrapped(self, *args, **kwargs):
        start = time()
        r = await func(self, *args, **kwargs)
        self.log.debug("%s took %s ms", func.__name__, round((time() - start) * 1000))
        return r

    return wrapped


def strip_leading_emoji(text: str) -> str:
    if not EMOJI_LIB_AVAILABLE:
        return text
    words = text.split(" ")
    # is_emoji returns False for 🛷️ for obscure reasons,
    # purely_emoji seems better
    if len(words) > 1 and emoji.purely_emoji(words[0]):
        return " ".join(words[1:])
    return text


async def noop_coro() -> None:
    pass


def add_quote_prefix(text: str):
    """
    Return multi-line text with leading quote marks (i.e. the ">" character).
    """
    return "\n".join(("> " + x).strip() for x in text.split("\n")).strip()



import pytest
from slixmpp import ComponentXMPP
from slixmpp.plugins.xep_0050.adhoc import XEP_0050

import slidge.command.adhoc
import slidge.command.base
from slidge.command import Command, CommandAccess
from slidge.command.adhoc import AdhocProvider
from slidge.command.base import Confirmation
from slidge.util.test import SlixTestPlus


class MockSession:
    def __init__(self, jid):
        self.logged = "logged" in jid.user


@pytest.fixture(autouse=True)
def mock(monkeypatch, MockRE):
    monkeypatch.setattr(
        slidge.command.base, "is_admin", lambda j: j.user.startswith("admin")
    )
    monkeypatch.setattr(Command, "_get_session", lambda s, j: MockSession(j))
    monkeypatch.setattr(XEP_0050, "new_session", lambda _: "session-id")
    monkeypatch.setattr(
        ComponentXMPP,
        "jid_validator",
        MockRE,
        raising=False,
    )
    monkeypatch.setattr(
        ComponentXMPP,
        "get_session_from_stanza",
        lambda s, j: MockSession(j.get_from()),
        raising=False,
    )
    monkeypatch.setattr(
        ComponentXMPP,
        "get_session_from_jid",
        lambda s, j: MockSession(j),
        raising=False,
    )


class Command1(Command):
    NAME = "Command number one"
    NODE = "command1"
    ACCESS = CommandAccess.ADMIN_ONLY


class Command2(Command1):
    NAME = "Command number two"
    NODE = "command2"
    ACCESS = CommandAccess.ADMIN_ONLY


class Command3(Command):
    NAME = "Command number three"
    NODE = "command3"

    CATEGORY = "category"
    ACCESS = CommandAccess.ADMIN_ONLY

    async def run(self, _session, _ifrom):
        return Confirmation(
            prompt="Confirm?", handler=self.finish, success="It worked!"
        )

    async def finish(self, _session, _ifrom):
        pass


class Command4(Command):
    NAME = "Command number four"
    NODE = "command4"

    CATEGORY = "category"
    ACCESS = CommandAccess.ADMIN_ONLY

    async def run(self, _session, _ifrom):
        return Confirmation(
            prompt="Confirm?", handler=self.finish, success="It worked!"
        )

    async def finish(self, _session, _ifrom):
        return "OK"


class TestCommandsDisco(SlixTestPlus):
    def setUp(self):
        self.stream_start(
            mode="component",
            plugins=["xep_0050"],
            jid="slidge.whatever.ass",
            server="whatever.ass",
        )
        self.adhoc = AdhocProvider(self.xmpp)
        self.adhoc.register(Command1(self.xmpp))
        self.adhoc.register(Command2(self.xmpp))
        self.adhoc.register(Command3(self.xmpp))
        self.adhoc.register(Command4(self.xmpp))
        super().setUp()

    def test_disco_admin(self):
        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <query xmlns="http://jabber.org/protocol/disco#items"
                     node="http://jabber.org/protocol/commands">
                <item jid="slidge.whatever.ass"
                      node="command1"
                      name="Command number one" />
                <item jid="slidge.whatever.ass"
                      node="command2"
                      name="Command number two" />
                <item jid="slidge.whatever.ass"
                      node="category"
                      name="category" />
              </query>
            </iq>
            """
        )

        self.recv(  # language=XML
            f"""
            <iq type='get'
                from='user@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'>
              <query xmlns='http://jabber.org/protocol/disco#items'
                     node='http://jabber.org/protocol/commands' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="user@whatever.ass/cheogram"
                id="2">
              <query xmlns="http://jabber.org/protocol/disco#items"
                     node="http://jabber.org/protocol/commands" />
            </iq>
            """
        )

    def test_non_existing_command(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='non-existing'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <error type="cancel">
                <item-not-found xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
              </error>
            </iq>
            """,
            use_values=False,
        )

    def test_category(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='category'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="category"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>category</title>
                  <field var="command"
                         type="list-single"
                         label="Command">
                    <option label="Command number three">
                      <value>command3</value>
                    </option>
                    <option label="Command number four">
                      <value>command4</value>
                    </option>
                    <value />
                  </field>
                </x>
              </command>
            </iq>
            """,
            use_values=False,
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='category'
                       sessionid="session-id">
                <x xmlns='jabber:x:data'
                   type='submit'>
                  <field var='command'>
                    <value>command3</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="category"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>Confirm?</title>
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from __future__ import unicode_literals

from typing import ClassVar, Set

from slixmpp.xmlstream import ElementBase


class Register(ElementBase):
    namespace = "jabber:iq:register"
    name = "query"
    plugin_attrib = "register"
    interfaces = {
        "username",
        "password",
        "email",
        "nick",
        "name",
        "first",
        "last",
        "address",
        "city",
        "state",
        "zip",
        "phone",
        "url",
        "date",
        "misc",
        "text",
        "key",
        "registered",
        "remove",
        "instructions",
        "fields",
    }
    sub_interfaces = interfaces
    form_fields = {
        "username",
        "password",
        "email",
        "nick",
        "name",
        "first",
        "last",
        "address",
        "city",
        "state",
        "zip",
        "phone",
        "url",
        "date",
        "misc",
        "text",
        "key",
    }

    def get_registered(self) -> bool:
        present = self.xml.find("{%s}registered" % self.namespace)
        return present is not None

    def get_remove(self) -> bool:
        present = self.xml.find("{%s}remove" % self.namespace)
        return present is not None

    def set_registered(self, value) -> None:
        if value:
            self.add_field("registered")
        else:
            del self["registered"]

    def set_remove(self, value) -> None:
        if value:
            self.add_field("remove")
        else:
            del self["remove"]

    def add_field(self, value: str) -> None:
        self._set_sub_text(value, "", keep=True)

    def get_fields(self):
        fields = set()
        for field in self.form_fields:
            if self.xml.find("{%s}%s" % (self.namespace, field)) is not None:
                fields.add(field)
        return fields

    def set_fields(self, fields) -> None:
        del self["fields"]
        for field in fields:
            self._set_sub_text(field, "", keep=True)

    def del_fields(self) -> None:
        for field in self.form_fields:
            self._del_sub(field)


class RegisterFeature(ElementBase):
    name = "register"
    namespace = "http://jabber.org/features/iq-register"
    plugin_attrib = name
    interfaces: ClassVar[Set[str]] = set()



"""
Slidge can be configured via CLI args, environment variables and/or INI files.

To use env vars, use this convention: ``--home-dir`` becomes ``HOME_DIR``.

Everything in ``/etc/slidge/conf.d/*`` is automatically used.
To use a plugin-specific INI file, put it in another dir,
and launch slidge with ``-c /path/to/plugin-specific.conf``.
Use the long version of the CLI arg without the double dash prefix inside this
INI file, eg ``debug=true``.

An example configuration file is available at
https://codeberg.org/slidge/slidge/src/branch/main/dev/confs/slidge-example.ini
"""

import asyncio
import importlib
import inspect
import logging
import logging.config
import os
import signal
from pathlib import Path

import configargparse

import slidge
from slidge import BaseGateway
from slidge.core import config
from slidge.db import SlidgeStore
from slidge.db.avatar import avatar_cache
from slidge.db.meta import get_engine
from slidge.migration import migrate
from slidge.util.conf import ConfigModule


class MainConfig(ConfigModule):
    def update_dynamic_defaults(self, args: configargparse.Namespace) -> None:
        # force=True is needed in case we call a logger before this is reached,
        # or basicConfig has no effect
        if args.log_config:
            logging.config.fileConfig(args.log_config)
        else:
            logging.basicConfig(
                level=args.loglevel,
                filename=args.log_file,
                force=True,
                format=args.log_format,
            )

        if args.home_dir is None:
            args.home_dir = Path("/var/lib/slidge") / str(args.jid)

        if args.db_url is None:
            args.db_url = f"sqlite:///{args.home_dir}/slidge.sqlite"


class SigTermInterrupt(Exception):
    pass


def get_configurator(from_entrypoint: bool = False) -> MainConfig:
    p = configargparse.ArgumentParser(
        default_config_files=[
            f"{p}/*"
            for p in os.getenv("SLIDGE_CONF_DIR", "/etc/slidge/conf.d/").split(":")
        ],
        description=__doc__,
    )
    p.add_argument(
        "-c",
        "--config",
        help="Path to a INI config file.",
        env_var="SLIDGE_CONFIG",
        is_config_file=True,
    )
    p.add_argument(
        "--log-config",
        help="Path to a INI config file to personalise logging output. Refer to "
        "<https://docs.python.org/3/library/logging.config.html#configuration-file-format> "
        "for details.",
    )
    p.add_argument(
        "-q",
        "--quiet",
        help="loglevel=WARNING (unused if --log-config is specified)",
        action="store_const",
        dest="loglevel",
        const=logging.WARNING,
        default=logging.INFO,
        env_var="SLIDGE_QUIET",
    )
    p.add_argument(
        "-d",
        "--debug",
        help="loglevel=DEBUG (unused if --log-config is specified)",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        env_var="SLIDGE_DEBUG",
    )
    p.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {slidge.__version__}",
    )
    configurator = MainConfig(
        config, p, skip_options=("legacy_module",) if from_entrypoint else ()
    )
    return configurator


def get_parser() -> configargparse.ArgumentParser:
    return get_configurator().parser


def configure(from_entrypoint: bool) -> list[str]:
    configurator = get_configurator(from_entrypoint)
    args, unknown_argv = configurator.set_conf()

    if not (h := config.HOME_DIR).exists():
        logging.info("Creating directory '%s'", h)
        os.makedirs(h)

    config.UPLOAD_REQUESTER = config.UPLOAD_REQUESTER or config.JID.bare

    return unknown_argv


def handle_sigterm(_signum: int, _frame) -> None:
    logging.info("Caught SIGTERM")
    raise SigTermInterrupt


def main(module_name: str | None = None) -> None:
    from_entrypoint = module_name is not None
    signal.signal(signal.SIGTERM, handle_sigterm)

    unknown_argv = configure(from_entrypoint)
    logging.info("Starting slidge version %s", slidge.__version__)

    if module_name is not None:
        config.LEGACY_MODULE = module_name

    legacy_module = importlib.import_module(config.LEGACY_MODULE)
    logging.debug("Legacy module: %s", dir(legacy_module))
    logging.info(
        "Starting legacy module: '%s' version %s",
        config.LEGACY_MODULE,
        getattr(legacy_module, "__version__", "No version"),
    )

    if plugin_config_obj := getattr(
        legacy_module, "config", getattr(legacy_module, "Config", None)
    ):
        # If the legacy module has default parameters that depend on dynamic defaults
        # of the slidge main config, it needs to be refreshed at this point, because
        # now the dynamic defaults are set.
        if inspect.ismodule(plugin_config_obj):
            importlib.reload(plugin_config_obj)
        logging.debug("Found a config object in plugin: %r", plugin_config_obj)
        ConfigModule.ENV_VAR_PREFIX += (
            f"_{config.LEGACY_MODULE.split('.')[-1].upper()}_"
        )
        logging.debug("Env var prefix: %s", ConfigModule.ENV_VAR_PREFIX)
        _, unknown_argv = ConfigModule(plugin_config_obj).set_conf(unknown_argv)

    if unknown_argv:
        logging.error(
            f"These config options have not been recognized and ignored: {unknown_argv}"
        )

    migrate()

    gw_cls: type[BaseGateway] = BaseGateway.get_unique_subclass()  # type:ignore[assignment]
    store = SlidgeStore(
        get_engine(
            config.DB_URL,
            echo=logging.getLogger().isEnabledFor(level=logging.DEBUG),
            pool_size=gw_cls.DB_POOL_SIZE,
        )
    )
    BaseGateway.store = store
    gateway = gw_cls()
    avatar_cache.store = gateway.store.avatars
    avatar_cache.set_dir(config.HOME_DIR / "slidge_avatars_v3")

    gateway.add_event_handler("disconnected", _on_disconnected)
    gateway.add_event_handler("stream_error", _on_stream_error)
    gateway.connect()
    return_code = 0
    try:
        gateway.loop.run_forever()
    except KeyboardInterrupt:
        logging.debug("Received SIGINT")
    except SigTermInterrupt:
        logging.debug("Received SIGTERM")
    except SystemExit as e:
        return_code = e.code  # type: ignore
        logging.debug("Exit called")
    except Exception as e:
        return_code = 2
        logging.exception("Exception in __main__")
        logging.exception(e)
    finally:
        if gateway.has_crashed:
            if return_code != 0:
                logging.warning("Return code has been set twice. Please report this.")
            return_code = 3
        if gateway.is_connected():
            logging.debug("Gateway is connected, cleaning up")
            gateway.del_event_handler("disconnected", _on_disconnected)
            gateway.loop.run_until_complete(asyncio.gather(*gateway.shutdown()))
            gateway.disconnect()
            gateway.loop.run_until_complete(gateway.disconnected)
            logging.info("Successful clean shut down")
        else:
            logging.debug("Gateway is not connected, no need to clean up")
        avatar_cache.close()
        gateway.loop.run_until_complete(gateway.http.close())
    logging.debug("Exiting with code %s", return_code)
    exit(return_code)


def _on_disconnected(e):
    logging.error("Disconnected from the XMPP server: '%s'.", e)
    exit(10)


def _on_stream_error(e):
    logging.error("Stream error: '%s'.", e)



import asyncio
import hashlib
import io
import logging
from concurrent.futures import ThreadPoolExecutor
from http import HTTPStatus
from pathlib import Path

import aiohttp
from multidict import CIMultiDictProxy
from PIL.Image import Image
from PIL.Image import open as open_image
from sqlalchemy import select

from ..core import config
from ..util.lock import NamedLockMixin
from ..util.types import Avatar as AvatarType
from .models import Avatar
from .store import AvatarStore


class CachedAvatar:
    def __init__(self, stored: Avatar, root_dir: Path) -> None:
        self.stored = stored
        self._root = root_dir

    @property
    def pk(self) -> int | None:
        return self.stored.id

    @property
    def hash(self) -> str:
        return self.stored.hash

    @property
    def height(self) -> int:
        return self.stored.height

    @property
    def width(self) -> int:
        return self.stored.width

    @property
    def etag(self) -> str | None:
        return self.stored.etag

    @property
    def last_modified(self) -> str | None:
        return self.stored.last_modified

    @property
    def data(self):
        return self.path.read_bytes()

    @property
    def path(self):
        return (self._root / self.hash).with_suffix(".png")


class NotModified(Exception):
    pass


class AvatarCache(NamedLockMixin):
    dir: Path
    http: aiohttp.ClientSession
    store: AvatarStore

    def __init__(self) -> None:
        self._thread_pool = ThreadPoolExecutor(config.AVATAR_RESAMPLING_THREADS)
        super().__init__()

    def get(self, stored: Avatar) -> CachedAvatar:
        return CachedAvatar(stored, self.dir)

    def set_dir(self, path: Path) -> None:
        self.dir = path
        self.dir.mkdir(exist_ok=True)
        log.debug("Checking avatar files")
        with self.store.session(expire_on_commit=False) as orm:
            for stored in orm.query(Avatar).all():
                avatar = CachedAvatar(stored, path)
                if avatar.path.exists():
                    continue
                log.warning(
                    "Removing avatar %s from store because %s does not exist",
                    avatar.hash,
                    avatar.path,
                )
                orm.delete(stored)
            orm.commit()

    def close(self) -> None:
        self._thread_pool.shutdown(cancel_futures=True)

    def __get_http_headers(self, cached: CachedAvatar | Avatar | None = None):
        headers = {}
        if cached and (self.dir / cached.hash).with_suffix(".png").exists():
            if last_modified := cached.last_modified:
                headers["If-Modified-Since"] = last_modified
            if etag := cached.etag:
                headers["If-None-Match"] = etag
        return headers

    async def __download(
        self,
        url: str,
        headers: dict[str, str],
    ) -> tuple[Image, CIMultiDictProxy[str]]:
        async with self.http.get(url, headers=headers) as response:
            if response.status == HTTPStatus.NOT_MODIFIED:
                log.debug("Using avatar cache for %s", url)
                raise NotModified
            response.raise_for_status()
            return (
                open_image(io.BytesIO(await response.read())),
                response.headers,
            )

    async def __is_modified(self, url, headers) -> bool:
        async with self.http.head(url, headers=headers) as response:
            return response.status != HTTPStatus.NOT_MODIFIED

    async def url_modified(self, url: str) -> bool:
        with self.store.session() as orm:
            cached = orm.query(Avatar).filter_by(url=url).one_or_none()
        if cached is None:
            return True
        headers = self.__get_http_headers(cached)
        return await self.__is_modified(url, headers)

    @staticmethod
    async def _get_image(avatar: AvatarType) -> Image:
        if avatar.data is not None:
            return open_image(io.BytesIO(avatar.data))
        elif avatar.path is not None:
            return open_image(avatar.path)
        raise TypeError("Avatar must be bytes or a Path", avatar)

    async def convert_or_get(self, avatar: AvatarType) -> CachedAvatar:
        if avatar.unique_id is not None:
            with self.store.session() as orm:
                stored = (
                    orm.query(Avatar)
                    .filter_by(legacy_id=str(avatar.unique_id))
                    .one_or_none()
                )
                if stored is not None:
                    return self.get(stored)

        if avatar.url is not None:
            return await self.__convert_url(avatar)

        return await self.convert(avatar, await self._get_image(avatar))

    async def __convert_url(self, avatar: AvatarType) -> CachedAvatar:
        assert avatar.url is not None
        async with self.lock(avatar.unique_id or avatar.url):
            with self.store.session() as orm:
                if avatar.unique_id is None:
                    stored = orm.query(Avatar).filter_by(url=avatar.url).one_or_none()
                else:
                    stored = (
                        orm.query(Avatar)
                        .filter_by(legacy_id=str(avatar.unique_id))
                        .one_or_none()
                    )
                    if stored is not None:
                        return self.get(stored)

            try:
                img, response_headers = await self.__download(
                    avatar.url, self.__get_http_headers(stored)
                )
            except NotModified:
                assert stored is not None
                return self.get(stored)

            return await self.convert(avatar, img, response_headers)

    async def convert(
        self,
        avatar: AvatarType,
        img: Image,
        response_headers: CIMultiDictProxy[str] | None = None,
    ) -> CachedAvatar:
        resize = (size := config.AVATAR_SIZE) and any(x > size for x in img.size)
        if resize:
            await asyncio.get_event_loop().run_in_executor(
                self._thread_pool, img.thumbnail, (size, size)
            )
            log.debug("Resampled image to %s", img.size)

        if (
            not resize
            and img.format == "PNG"
            and avatar.path is not None
            and avatar.path.exists()
        ):
            img_bytes = avatar.path.read_bytes()
        else:
            with io.BytesIO() as f:
                img.save(f, format="PNG")
                img_bytes = f.getvalue()

        hash_ = hashlib.sha1(img_bytes).hexdigest()
        file_path = (self.dir / hash_).with_suffix(".png")
        if file_path.exists():
            log.warning("Overwriting %s", file_path)
        with file_path.open("wb") as file:
            file.write(img_bytes)

        with self.store.session(expire_on_commit=False) as orm:
            stored = orm.execute(select(Avatar).where(Avatar.hash == hash_)).scalar()

            if stored is not None:
                if avatar.unique_id is not None:
                    if str(avatar.unique_id) != stored.legacy_id:
                        log.warning(
                            "Updating the 'unique' ID of an avatar, was '%s', is now '%s'",
                            stored.legacy_id,
                            avatar.unique_id,
                        )
                        stored.legacy_id = str(avatar.unique_id)
                        orm.add(stored)
                        orm.commit()

                return self.get(stored)

        stored = Avatar(
            hash=hash_,
            height=img.height,
            width=img.width,
            url=avatar.url,
            legacy_id=avatar.unique_id,
        )
        if response_headers:
            stored.etag = response_headers.get("etag")
            stored.last_modified = response_headers.get("last-modified")

        with self.store.session(expire_on_commit=False) as orm:
            if avatar.url is not None:
                existing = orm.execute(
                    select(Avatar).filter_by(url=avatar.url)
                ).scalar_one_or_none()
                if existing is not None:
                    orm.delete(existing)
                    orm.commit()
            orm.add(stored)
            orm.commit()
            return self.get(stored)


avatar_cache = AvatarCache()
log = logging.getLogger(__name__)
_download_lock = asyncio.Lock()

__all__ = (
    "CachedAvatar",
    "avatar_cache",
)



import logging
import uuid
import warnings
from typing import TYPE_CHECKING

from slixmpp import Iq, Message
from slixmpp.plugins.xep_0004 import Form

from ...util.types import ChatState, LegacyMessageType, Marker
from .attachment import AttachmentMixin
from .message_maker import MessageMaker
from .message_text import TextMessageMixin

if TYPE_CHECKING:
    from ...group import LegacyMUC

# this is for MDS
PUBLISH_OPTIONS = Form()
PUBLISH_OPTIONS["type"] = "submit"
PUBLISH_OPTIONS.add_field(
    "FORM_TYPE", "hidden", value="http://jabber.org/protocol/pubsub#publish-options"
)
PUBLISH_OPTIONS.add_field("pubsub#persist_items", value="true")
PUBLISH_OPTIONS.add_field("pubsub#max_items", value="max")
PUBLISH_OPTIONS.add_field("pubsub#send_last_published_item", value="never")
PUBLISH_OPTIONS.add_field("pubsub#access_model", value="whitelist")


class ChatStateMixin(MessageMaker):
    def __init__(self) -> None:
        super().__init__()
        self.__last_chat_state: ChatState | None = None

    def _chat_state(self, state: ChatState, forced: bool = False, **kwargs) -> None:
        carbon = kwargs.get("carbon", False)
        if carbon or (state == self.__last_chat_state and not forced):
            return
        self.__last_chat_state = state
        msg = self._make_message(state=state, hints={"no-store"})
        self._send(msg, **kwargs)

    def active(self, **kwargs) -> None:
        """
        Send an "active" chat state (:xep:`0085`) from this
        :term:`XMPP Entity`.
        """
        self._chat_state("active", **kwargs)

    def composing(self, **kwargs) -> None:
        """
        Send a "composing" (ie "typing notification") chat state (:xep:`0085`)
        from this :term:`XMPP Entity`.
        """
        self._chat_state("composing", forced=True, **kwargs)

    def paused(self, **kwargs) -> None:
        """
        Send a "paused" (ie "typing paused notification") chat state
        (:xep:`0085`) from this :term:`XMPP Entity`.
        """
        self._chat_state("paused", **kwargs)

    def inactive(self, **kwargs) -> None:
        """
        Send an "inactive" (ie "contact has not interacted with the chat session
        interface for an intermediate period of time") chat state (:xep:`0085`)
        from this :term:`XMPP Entity`.
        """
        self._chat_state("inactive", **kwargs)

    def gone(self, **kwargs) -> None:
        """
        Send a "gone" (ie "contact has not interacted with the chat session interface,
        system, or device for a relatively long period of time") chat state
        (:xep:`0085`) from this :term:`XMPP Entity`.
        """
        self._chat_state("gone", **kwargs)


class MarkerMixin(MessageMaker):
    def _make_marker(
        self, legacy_msg_id: LegacyMessageType, marker: Marker, carbon: bool = False
    ):
        msg = self._make_message(carbon=carbon)
        msg[marker]["id"] = self._legacy_to_xmpp(legacy_msg_id)[-1]
        return msg

    def ack(self, legacy_msg_id: LegacyMessageType, **kwargs) -> None:
        """
        Send an "acknowledged" message marker (:xep:`0333`) from this :term:`XMPP Entity`.

        :param legacy_msg_id: The message this marker refers to
        """
        self._send(
            self._make_marker(
                legacy_msg_id, "acknowledged", carbon=bool(kwargs.get("carbon"))
            ),
            **kwargs,
        )

    def received(self, legacy_msg_id: LegacyMessageType, **kwargs) -> None:
        """
        Send a "received" message marker (:xep:`0333`) from this :term:`XMPP Entity`.
        If called on a :class:`LegacyContact`, also send a delivery receipt
        marker (:xep:`0184`).

        :param legacy_msg_id: The message this marker refers to
        """
        carbon = bool(kwargs.get("carbon"))
        if self.mtype == "chat":
            for msg_id in self._legacy_to_xmpp(legacy_msg_id):
                self._send(
                    self.xmpp.delivery_receipt.make_ack(
                        msg_id,
                        mfrom=self.jid,
                        mto=self.user_jid,
                    )
                )
        self._send(
            self._make_marker(legacy_msg_id, "received", carbon=carbon), **kwargs
        )

    def displayed(self, legacy_msg_id: LegacyMessageType, **kwargs) -> None:
        """
        Send a "displayed" message marker (:xep:`0333`) from this :term:`XMPP Entity`.

        :param legacy_msg_id: The message this marker refers to
        """
        if (
            self.xmpp.MARK_ALL_MESSAGES
            and (muc := getattr(self, "muc", None))
            and getattr(self, "is_user", False)
        ):
            muc_jid = muc.jid.bare
            with self.xmpp.store.session() as orm:
                if self.xmpp.store.mam.is_displayed_by_user(
                    orm, muc_jid, str(legacy_msg_id)
                ):
                    self.session.log.debug(
                        "Ignoring carbon marker for message already displayed by user."
                    )
                    return
                else:
                    muc.pop_unread_xmpp_ids_up_to(
                        self._legacy_to_xmpp(legacy_msg_id)[-1]
                    )

        self._send(
            self._make_marker(
                legacy_msg_id, "displayed", carbon=bool(kwargs.get("carbon"))
            ),
            **kwargs,
        )
        if getattr(self, "is_user", False):
            self.session.create_task(self.__send_mds(legacy_msg_id))

    async def __send_mds(self, legacy_msg_id: LegacyMessageType) -> None:
        # Send a MDS displayed marker on behalf of the user for a group chat
        if muc := getattr(self, "muc", None):
            muc_jid = muc.jid.bare
        else:
            # This is not implemented for 1:1 chat because it would rely on
            # storing the XMPP-server injected stanza-id, which we don't track
            # ATM.
            # In practice, MDS should mostly be useful for public group chats,
            # so it should not be an issue.
            # We'll see if we need to implement that later
            return
        xmpp_msg_id = self._legacy_to_xmpp(legacy_msg_id)[-1]
        iq = Iq(sto=self.user_jid.bare, sfrom=self.user_jid.bare, stype="set")
        iq["pubsub"]["publish"]["node"] = self.xmpp["xep_0490"].stanza.NS
        iq["pubsub"]["publish"]["item"]["id"] = muc_jid
        displayed = self.xmpp["xep_0490"].stanza.Displayed()
        displayed["stanza_id"]["id"] = xmpp_msg_id
        displayed["stanza_id"]["by"] = muc_jid
        iq["pubsub"]["publish"]["item"]["payload"] = displayed
        iq["pubsub"]["publish_options"] = PUBLISH_OPTIONS
        try:
            await self.xmpp["xep_0356"].send_privileged_iq(iq)
        except Exception as e:
            self.session.log.debug("Could not MDS mark", exc_info=e)


class ContentMessageMixin(AttachmentMixin, TextMessageMixin):
    pass


class CarbonMessageMixin(ContentMessageMixin, MarkerMixin):
    def _privileged_send(self, msg: Message) -> None:
        i = msg.get_id()
        if i:
            self.session.ignore_messages.add(i)
        else:
            i = "slidge-carbon-" + str(uuid.uuid4())
            msg.set_id(i)
        msg.del_origin_id()
        try:
            self.xmpp["xep_0356"].send_privileged_message(msg)
        except PermissionError:
            warnings.warn(
                "Slidge does not have the privilege (XEP-0356) to send messages on behalf of users. "
                "Consider configuring your XMPP server for that."
            )


class InviteMixin(MessageMaker):
    def invite_to(
        self,
        muc: "LegacyMUC",
        reason: str | None = None,
        password: str | None = None,
        **send_kwargs,
    ) -> None:
        """
        Send an invitation to join a group (:xep:`0249`) from this :term:`XMPP Entity`.

        :param muc: the muc the user is invited to
        :param reason: a text explaining why the user should join this muc
        :param password: maybe this will make sense later? not sure
        :param send_kwargs: additional kwargs to be passed to _send()
            (internal use by slidge)
        """
        msg = self._make_message(mtype="normal")
        msg["groupchat_invite"]["jid"] = muc.jid
        if reason:
            msg["groupchat_invite"]["reason"] = reason
        if password:
            msg["groupchat_invite"]["password"] = password
        self._send(msg, **send_kwargs)


class MessageMixin(InviteMixin, ChatStateMixin, MarkerMixin, ContentMessageMixin):
    pass


class MessageCarbonMixin(InviteMixin, ChatStateMixin, CarbonMessageMixin):
    pass


log = logging.getLogger(__name__)



from typing import NamedTuple

from .base import NODE_PREFIX


class CommandCategory(NamedTuple):
    name: str
    node: str


ADMINISTRATION = CommandCategory("🛷️ Slidge administration", NODE_PREFIX + "admin")
CONTACTS = CommandCategory("👤 Contacts", NODE_PREFIX + "contacts")
GROUPS = CommandCategory("👥 Groups", NODE_PREFIX + "groups")



# type:ignore
import os
import tempfile
import types
from pathlib import Path
from xml.dom.minidom import parseString

try:
    import xmldiff.main

    XML_DIFF_PRESENT = True
except ImportError:
    xmldiff = None
    XML_DIFF_PRESENT = False

from slixmpp import (
    JID,
    ElementBase,
    Iq,
    MatcherId,
    MatchXMLMask,
    MatchXPath,
    Message,
    Presence,
    StanzaPath,
)
from slixmpp.stanza.error import Error
from slixmpp.test import SlixTest, TestTransport
from slixmpp.xmlstream import highlight, tostring
from slixmpp.xmlstream.matcher import MatchIDSender
from sqlalchemy import create_engine, delete

from slidge import (
    BaseGateway,
    BaseSession,
    LegacyBookmarks,
    LegacyContact,
    LegacyMUC,
    LegacyParticipant,
    LegacyRoster,
)

from ..command import Command
from ..core import config
from ..core.session import _sessions
from ..db import SlidgeStore
from ..db.avatar import avatar_cache
from ..db.meta import Base
from ..db.models import Contact, GatewayUser
from ..util import SubclassableOnce


class SlixTestPlus(SlixTest):
    def setUp(self) -> None:
        super().setUp()
        Error.namespace = "jabber:component:accept"

    def check(
        self,
        stanza,
        criteria,
        method: str = "exact",
        defaults=None,
        use_values: bool = True,
    ):
        """
        Create and compare several stanza objects to a correct XML string.

        If use_values is False, tests using stanza.values will not be used.

        Some stanzas provide default values for some interfaces, but
        these defaults can be problematic for testing since they can easily
        be forgotten when supplying the XML string. A list of interfaces that
        use defaults may be provided and the generated stanzas will use the
        default values for those interfaces if needed.

        However, correcting the supplied XML is not possible for interfaces
        that add or remove XML elements. Only interfaces that map to XML
        attributes may be set using the defaults parameter. The supplied XML
        must take into account any extra elements that are included by default.

        Arguments:
            stanza       -- The stanza object to test.
            criteria     -- An expression the stanza must match against.
            method       -- The type of matching to use; one of:
                            'exact', 'mask', 'id', 'xpath', and 'stanzapath'.
                            Defaults to the value of self.match_method.
            defaults     -- A list of stanza interfaces that have default
                            values. These interfaces will be set to their
                            defaults for the given and generated stanzas to
                            prevent unexpected test failures.
            use_values   -- Indicates if testing using stanza.values should
                            be used. Defaults to True.
        """
        if method is None and hasattr(self, "match_method"):
            method = getattr(self, "match_method")

        if method != "exact":
            matchers = {
                "stanzapath": StanzaPath,
                "xpath": MatchXPath,
                "mask": MatchXMLMask,
                "idsender": MatchIDSender,
                "id": MatcherId,
            }
            Matcher = matchers.get(method, None)
            if Matcher is None:
                raise ValueError("Unknown matching method.")
            test = Matcher(criteria)
            self.assertTrue(
                test.match(stanza),
                f"Stanza did not match using {method} method:\n"
                + f"Criteria:\n{str(criteria)}\n"
                + f"Stanza:\n{str(stanza)}",
            )
        else:
            stanza_class = stanza.__class__
            # Hack to preserve namespaces instead of having jabber:client
            # everywhere.
            old_ns = stanza_class.namespace
            stanza_class.namespace = stanza.namespace
            if not isinstance(criteria, ElementBase):
                xml = self.parse_xml(criteria)
            else:
                xml = criteria.xml

            # Ensure that top level namespaces are used, even if they
            # were not provided.
            self.fix_namespaces(stanza.xml)
            self.fix_namespaces(xml)

            stanza2 = stanza_class(xml=xml)

            if use_values:
                # Using stanza.values will add XML for any interface that
                # has a default value. We need to set those defaults on
                # the existing stanzas and XML so that they will compare
                # correctly.
                default_stanza = stanza_class()
                if defaults is None:
                    known_defaults = {Message: ["type"], Presence: ["priority"]}
                    defaults = known_defaults.get(stanza_class, [])
                for interface in defaults:
                    stanza[interface] = stanza[interface]
                    stanza2[interface] = stanza2[interface]
                    # Can really only automatically add defaults for top
                    # level attribute values. Anything else must be accounted
                    # for in the provided XML string.
                    if interface not in xml.attrib:
                        if interface in default_stanza.xml.attrib:
                            value = default_stanza.xml.attrib[interface]
                            xml.attrib[interface] = value

                values = stanza2.values
                stanza3 = stanza_class()
                stanza3.values = values

                debug = "Three methods for creating stanzas do not match.\n"
                debug += f"Given XML:\n{highlight(tostring(xml))}\n"
                debug += f"Given stanza:\n{format_stanza(stanza)}\n"
                debug += f"Generated stanza:\n{highlight(tostring(stanza2.xml))}\n"
                debug += (
                    f"Second generated stanza:\n{highlight(tostring(stanza3.xml))}\n"
                )
                result = self.compare(xml, stanza.xml, stanza2.xml, stanza3.xml)
            else:
                debug = "Two methods for creating stanzas do not match.\n"
                debug += f"Given XML:\n{highlight(tostring(xml))}\n"
                debug += f"Given stanza:\n{format_stanza(stanza)}\n"
                debug += f"Generated stanza:\n{highlight(tostring(stanza2.xml))}\n"
                result = self.compare(xml, stanza.xml, stanza2.xml)
            stanza_class.namespace = old_ns

            if XML_DIFF_PRESENT and not result:
                debug += str(
                    xmldiff.main.diff_texts(tostring(xml), tostring(stanza.xml))
                )
                if use_values:
                    debug += str(
                        xmldiff.main.diff_texts(tostring(xml), tostring(stanza2.xml))
                    )
            self.assertTrue(result, debug)

    def next_sent(self, timeout: float = 0.05) -> Message | Iq | Presence | None:
        self.wait_for_send_queue()
        sent = self.xmpp.socket.next_sent(timeout=timeout)
        if sent is None:
            return None
        xml = self.parse_xml(sent)
        self.fix_namespaces(xml, "jabber:component:accept")
        sent = self.xmpp._build_stanza(xml, "jabber:component:accept")
        return sent


class SlidgeTest(SlixTestPlus):
    plugin: types.ModuleType | dict

    class Config:
        jid = "aim.shakespeare.lit"
        secret = "test"
        server = "shakespeare.lit"
        port = 5222
        upload_service = "upload.test"
        home_dir = Path(tempfile.mkdtemp())
        user_jid_validator = ".*"
        admins: list[str] = []
        upload_requester = None
        ignore_delay_threshold = 300

    @classmethod
    def setUpClass(cls) -> None:
        for k, v in vars(cls.Config).items():
            setattr(config, k.upper(), v)
        if hasattr(cls, "plugin"):
            subclasses = SubclassableOnce._SubclassableOnce__subclasses
            subclasses.update(
                {
                    BaseGateway: find_subclass(cls.plugin, BaseGateway),
                    BaseSession: find_subclass(cls.plugin, BaseSession),
                    LegacyRoster: find_subclass(cls.plugin, LegacyRoster, base_ok=True),
                    LegacyContact: find_subclass(
                        cls.plugin, LegacyContact, base_ok=True
                    ),
                    LegacyMUC: find_subclass(cls.plugin, LegacyMUC, base_ok=True),
                    LegacyBookmarks: find_subclass(
                        cls.plugin, LegacyBookmarks, base_ok=True
                    ),
                }
            )
            for subclass in subclasses.values():
                if subclass is not None:
                    subclass.__abstractmethods__ = set()

    def setUp(self):
        # workaround for duplicate output of sql alchemy's log, cf
        # https://stackoverflow.com/a/76498428/5902284
        from sqlalchemy import log as sqlalchemy_log
        from sqlalchemy.pool import StaticPool

        sqlalchemy_log._add_default_handler = lambda x: None
        db_url = os.getenv("SLIDGETEST_DB_URL", "sqlite+pysqlite:///:memory:")
        engine = self.db_engine = create_engine(
            db_url, poolclass=StaticPool if db_url.startswith("postgresql+") else None
        )
        Base.metadata.create_all(engine)
        BaseGateway.store = SlidgeStore(engine)
        BaseGateway._test_mode = True

        self.xmpp = BaseGateway.get_self_or_unique_subclass()()

        self.xmpp.TEST_MODE = True
        avatar_cache.store = self.xmpp.store.avatars
        avatar_cache.set_dir(Path(tempfile.mkdtemp()))
        self.xmpp._always_send_everything = True
        engine.echo = True

        self.xmpp.connection_made(TestTransport(self.xmpp))
        self.xmpp.session_bind_event.set()
        # Remove unique ID prefix to make it easier to test
        self.xmpp._id_prefix = ""
        self.xmpp.default_lang = None
        self.xmpp.peer_default_lang = None

        def new_id():
            self.xmpp._id += 1
            return str(self.xmpp._id)

        self.xmpp._id = 0
        self.xmpp.new_id = new_id

        # Must have the stream header ready for xmpp.process() to work.
        header = self.xmpp.stream_header

        self.xmpp.data_received(header)
        self.wait_for_send_queue()

        self.xmpp.socket.next_sent()
        self.xmpp.socket.next_sent()

        # Some plugins require messages to have ID values. Set
        # this to True in tests related to those plugins.
        self.xmpp.use_message_ids = False
        self.xmpp.use_presence_ids = False
        Error.namespace = "jabber:component:accept"

    def _add_slidge_user(
        self,
        jid: str | JID = "romeo@shakespeare.lit",
        legacy_module_data=None,
        preferences=None,
    ) -> None:
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID(jid),
                legacy_module_data=legacy_module_data or {},
                preferences=preferences
                or {"sync_avatar": False, "sync_presence": False},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(Iq(sfrom=JID(jid)))
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing groups" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert stanza["status"].lower()

    def user_session(self, jid: str | JID = "romeo@shakespeare.lit") -> BaseSession:
        return BaseSession.get_self_or_unique_subclass().from_jid(JID(jid))

    def get_joined_muc(
        self,
        legacy_id: str | int,
        user_jid: str | JID = "romeo@shakespeare.lit",
        resource: str = "gajim",
    ) -> LegacyMUC:
        muc: LegacyMUC = self.run_coro(
            self.user_session(user_jid).bookmarks.by_legacy_id(legacy_id)
        )
        muc.add_user_resource(resource)
        return muc

    def tearDown(self) -> None:
        self.db_engine.echo = False
        super().tearDown()
        Base.metadata.drop_all(self.xmpp.store._engine)
        self.db_engine.dispose()
        _sessions.clear()

    def setup_logged_session(self, n_contacts: int = 0) -> None:
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()

        with self.xmpp.store.session() as session:
            session.execute(delete(Contact))
            session.commit()

        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="romeo@montague.lit/gajim")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"], welcome
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        if BaseGateway.get_self_or_unique_subclass().GROUPS:
            stanza = self.next_sent()
            assert "syncing groups" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "yup" in stanza["status"].lower(), stanza
        self.romeo: BaseSession = BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

        self.juliet: LegacyContact = self.run_coro(
            self.romeo.contacts.by_legacy_id("juliet")
        )
        self.room: LegacyMUC = self.run_coro(self.romeo.bookmarks.by_legacy_id("room"))
        self.first_witch: LegacyParticipant = self.run_coro(
            self.room.get_participant("firstwitch")
        )
        self.send(  # language=XML
            """
            <iq type="get"
                to="romeo@montague.lit"
                id="1"
                from="aim.shakespeare.lit">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:avatar:metadata" />
              </pubsub>
            </iq>
            """
        )

    @classmethod
    def tearDownClass(cls) -> None:
        reset_subclasses()


def format_stanza(stanza):
    return highlight(
        "\n".join(parseString(tostring(stanza.xml)).toprettyxml().split("\n")[1:])
    )


def find_subclass(o, parent, base_ok: bool = False):
    try:
        vals = vars(o).values()
    except TypeError:
        vals = o.values()
    for x in vals:
        try:
            if issubclass(x, parent) and x is not parent:
                return x
        except TypeError:
            pass
    else:
        if base_ok:
            return parent
        else:
            raise RuntimeError


def reset_subclasses() -> None:
    """
    Reset registered subclasses between test classes.

    Needed because these classes are meant to only be subclassed once and raise
    exceptions otherwise.
    """
    BaseSession.reset_subclass()
    BaseGateway.reset_subclass()
    LegacyRoster.reset_subclass()
    LegacyContact.reset_subclass()
    LegacyMUC.reset_subclass()
    LegacyBookmarks.reset_subclass()
    LegacyParticipant.reset_subclass()
    # reset_commands()


def reset_commands() -> None:
    Command.subclasses = [
        c for c in Command.subclasses if str(c).startswith("<class 'slidge.core")
    ]



import pytest
from slixmpp import JID, ComponentXMPP
from slixmpp.plugins.xep_0050.adhoc import XEP_0050
from slixmpp.test import SlixTest

import slidge.command.adhoc
from slidge.command import Command, TableResult
from slidge.command.adhoc import AdhocProvider
from slidge.command.base import FormField


class MockSession:
    def __init__(self, jid):
        self.logged = True


@pytest.fixture(autouse=True)
def mock(monkeypatch, MockRE):
    monkeypatch.setattr(
        slidge.command.base, "is_admin", lambda j: j.user.startswith("admin")
    )
    monkeypatch.setattr(Command, "_get_session", lambda s, j: MockSession(j))
    monkeypatch.setattr(
        ComponentXMPP,
        "jid_validator",
        MockRE,
        raising=False,
    )
    monkeypatch.setattr(XEP_0050, "new_session", lambda _: "session-id")


class Command1(Command):
    NAME = "Command number one"
    NODE = "command1"

    async def run(self, _session, _ifrom):
        return TableResult(
            description="A description",
            fields=[
                FormField("name", label="JID"),
                FormField("jid", type="jid-single", label="JID"),
            ],
            items=[
                {"jid": JID("test@test"), "name": "Some dude"},
                {"jid": "test2@test", "name": "Some dude2"},
            ],
        )


class TestCommandsResults(SlixTest):
    def setUp(self):
        self.stream_start(
            mode="component",
            plugins=["xep_0050"],
            jid="slidge.whatever.ass",
            server="whatever.ass",
        )
        self.adhoc = AdhocProvider(self.xmpp)
        self.adhoc.register(Command1(self.xmpp))

    def test_table_result(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="completed">
                <x xmlns="jabber:x:data"
                   type="result">
                  <title>A description</title>
                  <reported>
                    <field var="name"
                           type="text-single"
                           label="JID" />
                    <field var="jid"
                           type="jid-single"
                           label="JID" />
                  </reported>
                  <item>
                    <field var="name">
                      <value>Some dude2</value>
                    </field>
                    <field var="jid">
                      <value>test2@test</value>
                    </field>
                  </item>
                  <item>
                    <field var="name">
                      <value>Some dude</value>
                    </field>
                    <field var="jid">
                      <value>test@test</value>
                    </field>
                  </item>
                </x>
              </command>
            </iq>
            """,
            use_values=False,
        )



import pytest
import slixmpp.test
from slixmpp import ComponentXMPP
from slixmpp.plugins.xep_0050.adhoc import XEP_0050

import slidge.command.adhoc
from slidge.command import Command, Confirmation
from slidge.command.adhoc import AdhocProvider
from slidge.util.test import SlixTestPlus


class MockSession:
    def __init__(self, jid):
        self.jid = jid
        self.logged = "logged" in jid.user


@pytest.fixture(autouse=True)
def mock(monkeypatch, MockRE):
    monkeypatch.setattr(
        slidge.command.base, "is_admin", lambda j: j.user.startswith("admin")
    )
    monkeypatch.setattr(Command, "_get_session", lambda s, j: MockSession(j))
    monkeypatch.setattr(
        slixmpp.test.ComponentXMPP,
        "get_session_from_stanza",
        lambda self, stanza: MockSession(stanza.get_from()),
        raising=False,
    )
    monkeypatch.setattr(XEP_0050, "new_session", lambda _: "session-id")
    monkeypatch.setattr(
        ComponentXMPP,
        "jid_validator",
        MockRE,
        raising=False,
    )


class CommandAdmin(Command):
    NAME = "Command number one"
    NODE = "command1"

    async def run(self, _session, _ifrom):
        return Confirmation(
            prompt="Confirm?", handler=self.finish, success="It worked!"
        )

    async def finish(self, _session, _ifrom):
        pass


class CommandAdminConfirmFail(CommandAdmin):
    NAME = "Command number two"
    NODE = "command2"

    async def run_admin(self):
        return Confirmation(
            prompt="Confirm?", handler=self.finish, success="It worked!"
        )

    async def finish(self, _session, _ifrom):
        raise RuntimeError("Ploup")


class TestCommandsConfirmation(SlixTestPlus):
    def setUp(self):
        super().setUp()
        self.stream_start(
            mode="component",
            plugins=["xep_0050"],
            jid="slidge.whatever.ass",
            server="whatever.ass",
        )
        self.adhoc = AdhocProvider(self.xmpp)
        self.adhoc.register(CommandAdmin(self.xmpp))
        self.adhoc.register(CommandAdminConfirmFail(self.xmpp))

    def test_confirmation_cancel(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>Confirm?</title>
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id"
                       action='cancel' />
            </iq>
            """
        )
        self.recv(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="canceled" />
            </iq>
            """
        )

    def test_confirmation_do_it(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>Confirm?</title>
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command1'
                       sessionid="session-id"
                       action='next'>
                <x xmlns="jabber:x:data"
                   type="submit">
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command1"
                       sessionid="session-id"
                       status="completed">
                <note type="info">It worked!</note>
              </command>
            </iq>
            """
        )

    def test_confirmation_fail(self):
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="1">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command2'
                       action='execute' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="result"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="1">
              <command xmlns="http://jabber.org/protocol/commands"
                       node="command2"
                       sessionid="session-id"
                       status="executing">
                <actions>
                  <next />
                </actions>
                <x xmlns="jabber:x:data"
                   type="form">
                  <title>Confirm?</title>
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.recv(  # language=XML
            f"""
            <iq type='set'
                from='admin@whatever.ass/cheogram'
                to='{self.xmpp.boundjid.bare}'
                id="2">
              <command xmlns='http://jabber.org/protocol/commands'
                       node='command2'
                       sessionid="session-id"
                       action='complete'>
                <x xmlns="jabber:x:data"
                   type="submit">
                  <field var="confirm"
                         label="Confirm"
                         type="boolean">
                    <value>1</value>
                  </field>
                </x>
              </command>
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                type="error"
                from="slidge.whatever.ass"
                to="admin@whatever.ass/cheogram"
                id="2">
              <error type="wait">
                <internal-server-error xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Ploup</text>
              </error>
            </iq>
            """,
            use_values=False,
        )



import asyncio
from typing import TYPE_CHECKING

from slixmpp import CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import StanzaBase

from ... import config
from ..util import DispatcherMixin, exceptions_to_xmpp_errors

if TYPE_CHECKING:
    from slidge.core.gateway import BaseGateway


class MamMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: "BaseGateway") -> None:
        super().__init__(xmpp)
        self.__mam_cleanup_task = xmpp.loop.create_task(self.__mam_cleanup())  # type:ignore[misc]
        xmpp.register_handler(
            CoroutineCallback(
                "MAM_query",
                StanzaPath("iq@type=set/mam"),
                self.__handle_mam,
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "MAM_get_from",
                StanzaPath("iq@type=get/mam"),
                self.__handle_mam_get_form,
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "MAM_get_meta",
                StanzaPath("iq@type=get/mam_metadata"),
                self.__handle_mam_metadata,
            )
        )

    async def __mam_cleanup(self) -> None:
        if not config.MAM_MAX_DAYS:
            return
        while True:
            await asyncio.sleep(3600 * 6)
            with self.xmpp.store.session() as orm:
                self.xmpp.store.mam.nuke_older_than(orm, config.MAM_MAX_DAYS)
                orm.commit()

    @exceptions_to_xmpp_errors
    async def __handle_mam(self, iq: Iq) -> None:
        muc = await self.get_muc_from_stanza(iq)
        await muc.send_mam(iq)

    async def __handle_mam_get_form(self, iq: StanzaBase):
        assert isinstance(iq, Iq)
        ito = iq.get_to()

        if ito == self.xmpp.boundjid.bare:
            raise XMPPError(
                text="No MAM on the component itself, use a JID with a resource"
            )

        session = await self._get_session(iq, 0, logged=True)
        await session.bookmarks.by_jid(ito)

        reply = iq.reply()
        form = self.xmpp.plugin["xep_0004"].make_form()
        form.add_field(ftype="hidden", var="FORM_TYPE", value="urn:xmpp:mam:2")
        form.add_field(ftype="jid-single", var="with")
        form.add_field(ftype="text-single", var="start")
        form.add_field(ftype="text-single", var="end")
        form.add_field(ftype="text-single", var="before-id")
        form.add_field(ftype="text-single", var="after-id")
        form.add_field(ftype="boolean", var="include-groupchat")
        field = form.add_field(ftype="list-multi", var="ids")
        field["validate"]["datatype"] = "xs:string"
        field["validate"]["open"] = True
        reply["mam"].append(form)
        reply.send()

    @exceptions_to_xmpp_errors
    async def __handle_mam_metadata(self, iq: Iq) -> None:
        muc = await self.get_muc_from_stanza(iq)
        await muc.send_mam_metadata(iq)



from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0030.stanza.info import DiscoInfo
from slixmpp.types import OptJid

from .base import Base


class BaseDiscoMixin(Base):
    DISCO_TYPE: str = NotImplemented
    DISCO_CATEGORY: str = NotImplemented
    DISCO_NAME: str = NotImplemented
    DISCO_LANG = None

    def _get_disco_name(self) -> str | None:
        if self.DISCO_NAME is NotImplemented:
            return self.xmpp.COMPONENT_NAME
        return self.DISCO_NAME or self.xmpp.COMPONENT_NAME

    def features(self):
        return []

    async def extended_features(self) -> list[Form] | None:
        return None

    async def get_disco_info(self, jid: OptJid = None, node: str | None = None):
        info = DiscoInfo()
        for feature in self.features():
            info.add_feature(feature)
        info.add_identity(
            category=self.DISCO_CATEGORY,
            itype=self.DISCO_TYPE,
            name=self._get_disco_name(),
            lang=self.DISCO_LANG,
        )
        if forms := await self.extended_features():
            for form in forms:
                info.append(form)
        return info

    async def get_caps_ver(self, jid: OptJid = None, node: str | None = None):
        info = await self.get_disco_info(jid, node)
        caps = self.xmpp.plugin["xep_0115"]
        ver = caps.generate_verstring(info, caps.hash)
        return ver


class ChatterDiscoMixin(BaseDiscoMixin):
    AVATAR = True
    RECEIPTS = True
    MARKS = True
    CHAT_STATES = True
    UPLOAD = True
    CORRECTION = True
    REACTION = True
    RETRACTION = True
    REPLIES = True
    INVITATION_RECIPIENT = False

    DISCO_TYPE = "pc"
    DISCO_CATEGORY = "client"
    DISCO_NAME = ""

    def features(self):
        features = []
        if self.CHAT_STATES:
            features.append("http://jabber.org/protocol/chatstates")
        if self.RECEIPTS:
            features.append("urn:xmpp:receipts")
        if self.CORRECTION:
            features.append("urn:xmpp:message-correct:0")
        if self.MARKS:
            features.append("urn:xmpp:chat-markers:0")
        if self.UPLOAD:
            features.append("jabber:x:oob")
        if self.REACTION:
            features.append("urn:xmpp:reactions:0")
        if self.RETRACTION:
            features.append("urn:xmpp:message-retract:0")
        if self.REPLIES:
            features.append("urn:xmpp:reply:0")
        if self.INVITATION_RECIPIENT:
            features.append("jabber:x:conference")
        features.append("urn:ietf:params:xml:ns:vcard-4.0")
        return features

    async def extended_features(self):
        f = getattr(self, "restricted_emoji_extended_feature", None)
        if f is None:
            return

        e = await f()
        if not e:
            return

        return [e]


class ContactAccountDiscoMixin(BaseDiscoMixin):
    async def get_disco_info(self, jid: OptJid = None, node: str | None = None):
        if jid and jid.resource:
            return await super().get_disco_info()
        info = DiscoInfo()
        info.add_feature("http://jabber.org/protocol/pubsub")
        info.add_feature("http://jabber.org/protocol/pubsub#retrieve-items")
        info.add_feature("http://jabber.org/protocol/pubsub#subscribe")
        info.add_identity(
            category="account",
            itype="registered",
            name=self._get_disco_name(),
            lang=self.DISCO_LANG,
        )
        info.add_identity(
            category="pubsub",
            itype="pep",
            name=self._get_disco_name(),
            lang=self.DISCO_LANG,
        )
        return info
